#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#define F_CPU 16000000UL

#include <util/delay.h>

// The frequency output by the speaker for comparison with the received frequency
#define SPKRFREQ 24000

// Parameters used for tuning the detection algorithm
#define THRHIGH 145      // Any value above this high threshold is taken as a "valid" (non-noise) value
#define SPIKETHR 70      // Threshold for a spike. If the increase between two peaks is more than 70, this action is immediately assigned as a PUSH
#define MINFORPP 1       // Minimum number of pushes and pulls. If the number of both pushes and pulls exceeds this number, then comparison will result in a decision (whether push or pull) made
#define WDELAY 30000     // Delay that the LED is kept on for to indicate an action. This is also used to ignore blips in amplitude. While this delay is counting down, further increases in amplitude are ignored.
#define CNTDN 2000       // Countdown to clear the buffer of peaks.
#define HSIZE 6          // Defines the size of the buffer of peaks. The higher this value, the more peaks over which a decision is made. The algorithm makes more mistakes when this number is small, but when it is too large, it may never be filled completely and detection will never be triggered

// motion detected
#define NONE 0
#define PUSH 1
#define PULL 2

// state definitions
#define FILL 0
#define CHECK 1
#define WAITNONE 2
#define WAITPUSH 3
#define FINDPEAK 4

// function definitions
void initialize(void);
void initSpkrGen(void);
void init_adc(void);
char detectMotion(unsigned char maxHighsPtr);

// buffer of peaks
unsigned char maxHighs[HSIZE] = { 0 };

int main(void) {
	initialize();
	
	unsigned char maxHighsPtr = 0;         // This variable creates a circular buffer out of the maxHighs array
	unsigned char maxHighsPtrChgd = 0;     // If set, it indicates that maxHighsPtr was incremented. Used to determine whether the state machine should transition to the CHECK state from the FINDPEAK state

	unsigned char currVal = 0;             // Used to temporarily hold the result of ADC sampling
	unsigned char prevVal = 0;             // Used to hold the current highest value sampled by the ADC (for peak detection)

	unsigned char tempMotion = NONE;       // Used to hold the result of the latest motion detection

	unsigned int wdelay = WDELAY;          // Initializes the delay for LEDs to be lit
	unsigned char state = FILL;            // Initializes the state of the state machine
	unsigned int countdown = CNTDN;        // Initializes the delay of the countdown for buffer clearing.

	unsigned char found = 0;               // If set, a peak has been found. Only start searching for the next peak (by setting found back to 0) when ADC samples next cross THRHIGH

	while(1){
	        // ADSC is first set in the initializer, this reads the last ADC sample into currVal
		currVal = ADCH;
		// This starts the ADC getting another sample
		ADCSRA |= (1 << ADSC);
		
		// This section searches for peaks
		if (currVal > THRHIGH && prevVal <= THRHIGH) { // If this is the first time the ADC sample passes THRHIGH, set found to 0 and set prevVal to currVal since currVal is the current highest ADC value
			found = 0; 
			prevVal = currVal;
		}
		
		if (currVal > THRHIGH && currVal > prevVal) { // If the ADC sample is valid (above THRHIGH) and this value is larger than the previous value, then prevVal is set to currVal
			prevVal = currVal;
		}
		else if (prevVal > currVal + 5 && !found) { // If the current value is lower than prevVal by 5 (this threshold of 5 is added to adjust for ADC inaccuracies), and not a peak has not been found, then add the peak to the peak buffer
                        // the ADC does not sample accurately. If a constant voltage were provided to the ADC, the ADC may still sample values that are one above or below due to rounding errors
		        // without the threshold of 5, currVal may just dip below prevVal by 1 due to rounding errors and cause prevVal to wrongly be picked as the peak.
			found = 1;
			maxHighs[maxHighsPtr] = prevVal;
			++maxHighsPtr;
			maxHighsPtrChgd = 1;
			countdown = CNTDN;
		}
		
		if (found && currVal <= 128) { // This resets prevVal to 0 after a peak is found and currVal dips below the 2.5 V point.
			prevVal = 0;
		}
		
		// This section is the state machine
		switch (state) {
			case (FILL): 
			        // This state fills the peak buffer with peaks

				// check if buffer is full
				if (maxHighsPtr == HSIZE) {
					state = CHECK;
					maxHighsPtr = 0;
				}

				// countdown to to clear the peak buffer
				countdown--;
				if (countdown == 0) {
					countdown = CNTDN;
					prevVal = 0;
					found = 0;
					maxHighsPtr = 0;
					PORTC |= (1<<PORTC0);
					PORTC &= ~(1<<PORTC1);
					PORTC &= ~(1<<PORTC2);
				}

				break;

			case (CHECK):
			        // This state checks to see if there was a motion detected based on the elements of the peak buffer

				tempMotion = detectMotion(maxHighsPtr);
				
				// transition states
				if (tempMotion == NONE) {
					state = WAITNONE;
					PORTC |= (1<<PORTC0);
					PORTC &= ~(1<<PORTC1);
					PORTC &= ~(1<<PORTC2);				
				}
				else if (tempMotion == PUSH) {
					state = WAITPUSH;
					wdelay = WDELAY;
					PORTC |= (1<<PORTC1);
					PORTC &= ~(1<<PORTC0);
					PORTC &= ~(1<<PORTC2);
				}
				else if (tempMotion == PULL) {
					state = WAITPUSH;
					wdelay = WDELAY;
					PORTC |= (1<<PORTC2);
					PORTC &= ~(1<<PORTC0);
					PORTC &= ~(1<<PORTC1);
				}
				break;

			case (WAITNONE):
				// This state is the result of no motion detected in the CHECK state. If the sampled ADC value is above THRHIGH, the next peak will be found

				if (currVal > THRHIGH) {
					state = FINDPEAK;
					prevVal = currVal;
				} else {
					// If the threshold is not exceeded, decrement counter. If timer times out, return to the fill state
					countdown--;
					if (countdown == 0) {
						countdown = CNTDN;
						state = FILL;
						prevVal = 0;
						maxHighsPtr = 0;
					}
				}
				break;

			case (WAITPUSH):
				// This state is the result of motion detected (both PUSHes and PULLs) in the CHECK state. After wdelay times out, the state machine returns to the FILL state.
				// While in this state, peaks are still recorded, but actually ignored since when the state machine returns to the FILL state, maxHighsPtr returns to 0 and the peak buffer is 'cleared'
				if (wdelay > 0) {
					--wdelay;
				} else {
					state = FILL;
					prevVal = 0;
					maxHighsPtr = 0;
				}
				break;

			case (FINDPEAK):
				// This state searches for a new peak if there was no motion detected in the WAITNONE state
				// This state is different from the FILL state in that if no motion is detected in the WAITNONE state, the window of peaks covered by the peak buffer may just be wrong or one-off. 
				// This state allows just one newest entry to replace the oldest entry in the peak buffer

				if (maxHighsPtrChgd) { // A new entry has just been added to the peak buffer, so CHECK again.
					state = CHECK;
				} else {
					// If nothing was changed, decrement timer, and if the timer times out, go to the FILL state 
					countdown--;
					if (countdown == 0) {
						countdown = CNTDN;
						prevVal = 0;
						found = 0;
						maxHighsPtr = 0;
						PORTC |= (1<<PORTC0);
						PORTC &= ~(1<<PORTC1);
						PORTC &= ~(1<<PORTC2);
						state = FILL;
					}
				}

				break;	
		}

		// This resetting back to 0 completes the circular buffer
		// We could have used %HSIZE when incrementing maxHighsPtr, but we think this is faster
		if (maxHighsPtr == HSIZE) {
			maxHighsPtr = 0;
		}
		
		// Resets the changed status to 0
		maxHighsPtrChgd = 0;

		// Inserted this because ADC sampling is not perfect and a later voltage, though higher, 
		// may be sampled as a value lower than the earlier value. This delay corrects for that
		_delay_us(10);
		
		// Useful debug outputs that we decided to leave in the final submission.
		OCR2A = currVal;    // Allows us to observe what the sampled value of the ADC was (Port D7)
		OCR2B = state*50;   // Allows us to observe what state the state machine is in at any point in time (Port D6
	}
}

char detectMotion(unsigned char maxHighsPtr) {
	unsigned char i;          // Used as the array index for the peak buffer
	unsigned char cnt;        // Counts the number of comparisons left to be made
	unsigned char push = 0;   // Holds the number of pushes in this set of 10 samples in the peak buffer
	unsigned char pull = 0;   // Holds the number of pulls in this set of 10 samples in the peak buffer

	// Initializes the array index for the peak buffer
	i = maxHighsPtr;
	
	// The number of comparisons is always one less than the number of elements in the peak buffer.
	cnt = HSIZE-1;
	while (cnt > 0) {
		// This odd increment is used because the peak buffer is a circular buffer
		unsigned char nexti = i + 1;
		if (nexti == HSIZE) {
			nexti = 0;
		}
		
		// If any one sample is greater than the previous one by SPIKETHR, a push must have happened
		// This occurs if a fast push was captured by the ADC
		if (maxHighs[nexti] > maxHighs[i] && (maxHighs[nexti] - maxHighs[i]) >= SPIKETHR) {
			return PUSH;
		}
		
		// Assigns push to one pair of values in the peak buffer if the peaks are increasing and pulls if the peaks are decreasing
		// Does not assign anything if peaks are equal
		if (maxHighs[i] < maxHighs[nexti]) {
			push++;
			// If the first three comparisons all indicate pushes, it must be a push
			if (push > 2 && push == HSIZE-1-cnt) {
				return PUSH;
			}
		} else if (maxHighs[i] > maxHighs[nexti]) {
			pull++;
			// If the first three comparisons all indicate pulls, it must be a pull
			if (pull > 2 && pull == HSIZE-1-cnt) {
				return PULL;
			}
		}
		
		--cnt;
		++i;
	}

	// If all the comparisons result in pushes, return push
	if (push == HSIZE-1) {
		return PUSH;
	}
	
	// If all the comparisons result in pulls, return pull
	if (pull == HSIZE-1) {
		return PULL;
	}
	
	// If there are at least MINFORPP pushes and pulls, return push if there are more pushes than pulls.
	if (push > MINFORPP && pull > MINFORPP) {
		if (push >= pull) {
			return PUSH;
		} else {
			return PULL;
		}
	} 
	
	// no motion detected
	return NONE;
}

void initialize(void) {
	initSpkrGen();
	
	// Used to indicate no motion, pushes and pulls
	DDRC |= (1<<DDC0); // No motion
	DDRC |= (1<<DDC1); // Push
	DDRC |= (1<<DDC2); // Pull
	PORTC = (0<<PORTC1) | (0<<PORTC0) | (0<<PORTC2);

	// PWM output used for debugging but intentionally left here as it indicates useful information
	TCCR2A = (0<<COM2A0) | (1<<COM2A1) | (0<<COM2B0) | (1<<COM2B1) | (1<<WGM20) | (1<<WGM21);
	TCCR2B = 4; // divide by 64
	DDRD |= (1<<DDD6);
	DDRD |= (1<<DDD7);

	init_adc();

	sei();
}

// initializes the ADC
void init_adc(void) {
	//enable adc, adc start conversion bit, clock pre-scalar is 4 (divide by 16)
	ADCSRA = ((1 << ADEN) | (1 << ADSC)) + 4; 
	ADMUX = ((1 << REFS0) | (1 << ADLAR)) + 0; 	//left adjust
							//channel0 is connected to ADC
	ADCSRA |= (1 << ADSC);
}

void initSpkrGen(void) {
	// Will use timer 0B because timer 0B has the toggle function built in, without complications with WGM02
	TCCR0A = (0<<COM0B1) | (1<<COM0B0) | (1<<WGM01) | (0<<WGM00);
	TCCR0B = (0<<WGM02) | (0<<CS02) | (1<<CS01) | (0<<CS00);
	
	// OCR0B should be set to 41 => calculated frequency is 24390, measured frequency is 23.83kHz.
	//OCR0B = F_CPU/8/(SPKRFREQ*2);
	OCR0A = 41;

	// Enable Pin B4
	DDRB |= (1<<DDB4);
}
