#include <stdio.h>
#include <stdio.h>
#include <avr/io.h>
#include <avr/eeprom.h>
#include "uart.h"

#define F_CPU 16000000UL
#include <util/delay.h>

#define CROSSINGS 500
#define RECORDLEN 5000

// push/pull state machine states
#define WEAKPUSH 0
#define STRONGPUSH 1
#define WEAKPULL 2
#define STRONGPULL 3

// push/pull states
#define NONE 0
#define PUSH 1
#define PULL 2

// function definitions
void initialize(void);
void initSpkrGen(void);
void init_adc(void);

// UART file descriptor
// putchar and getchar are in uart.c
FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);

char record[RECORDLEN];

double crossings[CROSSINGS] = { 0.0 };
unsigned int crsIdx = 0;
unsigned char zero = 128; // zero on 8-bit ADC is 256/2

// Use point-slope to find exact zero crossing given two points on a line
double findCrossing(unsigned char y1, unsigned char y2, unsigned int x2) {
	unsigned int x1 = x2-1;
	double m = ((double)(x2-x1))/((double)(y2-y1));
	double c = (zero-y1)*(m)+x1;
	return c;
}

// Find the max value within a range from start to end
unsigned char findPeak(unsigned int *startPtr, unsigned int end) {
	unsigned int start = *startPtr; // start of range
	unsigned char prevVal = 0;
	unsigned char currVal;
	char hasSampled = 0; // true if localPeak has a positive value
	unsigned char localPeak = 0; // keep track of the current max
  // loop over the entire range for the max
	while (start < end) {
		currVal = record[start];
		// rising edge of the wave
		if (hasSampled == 0) {
      // if no peak value has been saved, and the peak is > "zero"
			if (currVal < prevVal && prevVal > 128) {
				hasSampled = 1;
				localPeak = prevVal;
			}
		}
		// falling edge of the wave
		else {
			if (currVal > prevVal) { 
				break; // break to return the max
			}
		}
		++start;
		prevVal = currVal;
	}
	*startPtr = start;
	return localPeak;
}

// break up a single block into one or more different motions
unsigned char blockMotion(unsigned int *startPtr, unsigned int end) {
	unsigned int start = *startPtr;
	unsigned char peak = 240; // define max peak @ 240 units
	unsigned char state = WEAKPUSH; // initial state
	unsigned char prevPeak = 0;
	unsigned char localPeak;
  // loop over entire range for a block section
	while (start < end) {
		localPeak = findPeak(&start, end);
		if (localPeak > peak) {
			break; // high amplitude indicates transition of gesture
		}
    // two-bit saturating counter state machine
		switch (state) {
      // weakpush can only transition to strong push or weak pull
			case (WEAKPUSH):
				if (localPeak > prevPeak) {
					state = STRONGPUSH;
				}
				else if (localPeak < prevPeak) {
					state = WEAKPULL;
				}
				break;
      // strongpush can only transition to weak push
			case (STRONGPUSH):
				if (localPeak < prevPeak) {
					state = WEAKPUSH;
				}	
				break;
      // weakpull can only transition to strong pull or weak push
			case (WEAKPULL):
				if (localPeak < prevPeak) {
					state = STRONGPULL;
				}
				else if (localPeak > prevPeak) {
					state = WEAKPUSH;
				}
				break;
      // strongpull can only transition to weak pull
			case (STRONGPULL):
				if (localPeak > prevPeak) {
					state = WEAKPULL;
				}
				break;
		}
		prevPeak = localPeak;
	}
	*startPtr = start;
	
  // weakpush and strongpush both evaluate to a push motion
	if (state == WEAKPUSH || state == STRONGPUSH) {
		return PUSH;
	}
  // else return pull
	return PULL;
}

// main method for the record and playback algorithm
int main() {
	initialize();
		
	// Give 3 s for user to get ready
	_delay_ms(3000);
	
	// Light up DDB2
	PORTB |= (1<<PORTB2);
	
	// Start recording
	unsigned int recordCnt = 0;		
	while (recordCnt < RECORDLEN) {
		record[recordCnt] = ADCH;
		ADCSRA |= (1 << ADSC);
		_delay_ms(1);
		recordCnt++;
	}
	
	// Switch off DDB2
	PORTB &= ~(1<<PORTB2);
	
	// Playback using PWM first as a visual check
	unsigned int playback = 0;
	while (playback < RECORDLEN) {
		OCR2A = record[playback];
		_delay_ms(1);
		playback++;
	}
	
  // Send all data to PC
	fprintf(stdout,"Data,");
	for(int i = 0; i < RECORDLEN; i++) {
		fprintf(stdout,"%d,", record[i]);
	}

	// find zero crossings
	unsigned char prevPoint = record[0];
	unsigned char currPoint;
	unsigned int i;
	for (i = 1; i < RECORDLEN; i++) {
		currPoint = record[i];
		if ((prevPoint < zero && currPoint >= zero) || (prevPoint > zero && currPoint <= zero)) {
			crossings[crsIdx++] = findCrossing(prevPoint, currPoint, i);
		}
		if (crsIdx == CROSSINGS) { break; }
		prevPoint = currPoint;
	}
	
	// find a block of push/pull
	unsigned char minBlock = 5; // minimum block size
	unsigned char minSamplesBet = 20; // < 20 samples between zero crossings
	unsigned int beginBlock = 0;
	char hasBlockBegun = 0;
	unsigned int crossingsInBlock = 0;
	unsigned int endBlock = 0;
	unsigned int prevCrossing = 0;
	unsigned int currCrossing;
	unsigned char state;
	unsigned char prevState = NONE;
	unsigned int tempBeginBlock = 0;
	unsigned int currBeginBlock = 0;
	unsigned int prevBeginBlock = 0;
	
	fprintf(stdout,"Motion,"); // indicates beginning of motion to python script

  // iterate over number of crossings to partition into blocks
	for (i = 1; i < crsIdx; i++) {
		currCrossing = crossings[i];
    // determine if crossing occurred at high frequency, and increment block size
		if (currCrossing < (prevCrossing+minSamplesBet)) {
			if (hasBlockBegun == 0) {
				beginBlock = prevCrossing;
				hasBlockBegun = 1;
			}
			crossingsInBlock++;
		}
    // once a low freq part has been reached, break up the last high freq part
		else {
      // check if the last block was high frequency
			if (crossingsInBlock > minBlock) {
				tempBeginBlock = beginBlock;
				endBlock = prevCrossing;
        // continue breaking up the block until the end has been reached
				while (tempBeginBlock != endBlock) {
					// break up block into motions
					state = blockMotion(&tempBeginBlock, endBlock);
          // only declare the last state if the current state is different (prevents multiple predictions)
					if (state != prevState) {
            // reset pointers if they have not been set
						if (prevBeginBlock == 0 && currBeginBlock == 0) {
							prevBeginBlock = beginBlock;
							currBeginBlock = beginBlock;
						}
            // indicates motion and duration to python script
						fprintf(stdout, "%d(%d|%d),", prevState, prevBeginBlock, currBeginBlock);
						// at the end of a block, if the state does not match, update block pointers
						if (currBeginBlock < beginBlock) {
							prevBeginBlock = beginBlock;
							currBeginBlock = tempBeginBlock;
						}
            // allow pointers to be updated in round-robin fashion
						else {
							prevBeginBlock = currBeginBlock;
							currBeginBlock = tempBeginBlock;
						}
					}
          // if the motion detected is the same, combine them into one detection rather than 2 small ones
					else {
						currBeginBlock = tempBeginBlock;
					}
					prevState = state;
				}
			}
			crossingsInBlock = 0;
			hasBlockBegun = 0;
		}
		prevCrossing = currCrossing;
	}
	// at the very end, print the last state
	fprintf(stdout, "%d(%d|%d)\r\n", prevState, prevBeginBlock, currBeginBlock);
	
  // loop at the end, program only runs once
	while (1) {
	
	}
}

// initialize all microcontroller components needed for program
void initialize(void) {
	initSpkrGen();
  
  // LED output for indicating recording period
	DDRB |= (1<<DDB0);	
	DDRB |= (1<<DDB1); 
	DDRB |= (1<<DDB2);
	PORTB = (0<<PORTB1) | (0<<PORTB0) | (0<<PORTB2);

	// output of waveform in playback
	TCCR2A = (0<<COM2A0) | (1<<COM2A1) | (0<<COM2B0) | (1<<COM2B1) | (1<<WGM20) | (1<<WGM21);
	TCCR2B = 4; // divide by 64
	DDRD |= (1<<DDD7);
	OCR2A = 0;

	init_adc();
	
  // initialize uart
	uart_init();
	stdout = stdin = stderr = &uart_str;
}

// initializes the ADC
void init_adc(void) {
	//enable adc, adc start conversion bit, clock pre-scalar is 4 (divide by 16)
	ADCSRA = ((1 << ADEN) | (1 << ADSC)) + 4; 
	ADMUX = ((1 << REFS0) | (1 << ADLAR)) + 0; 	//left adjust, channel0 is connected to ADC
	ADCSRA |= (1 << ADSC); // start adc
}

// initialize PWM at 23.8kHz for sound through speaker
void initSpkrGen(void) {

	// Will use timer 0B because timer 0B has the toggle function built in, without complications with WGM02
	TCCR0A = (0<<COM0B1) | (1<<COM0B0) | (1<<WGM01) | (0<<WGM00);
	TCCR0B = (0<<WGM02) | (0<<CS02) | (1<<CS01) | (0<<CS00);
	
	// OCR0B should be set to 41 => calculated frequency is 24390, measured frequency is 23.83kHz.
	//OCR0B = F_CPU/8/(SPKRFREQ*2);
	OCR0A = 41;

	// Enable Pin B4
	DDRB |= (1<<DDB4);
}


